#==============================================
#  RES203  Internet applications
#  TP Caching
#----------------------------------------------
# this code can be used standalone, 
# however juptyter makes it much more readable
# also, the text to guide you in the TP is not
# available as a comment...


import math
import numpy as np
import timeit

from scipy import integrate
from scipy.optimize import fsolve

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.colors import cnames
from matplotlib import animation
from itertools import takewhile



def solve_lfu(M=1e7, C=1e3, A=1.0, verbose=True):  
    """ 
    Compute LFU cache hit ratio
    """
    print("solve_lfu(%g,%s,%s)" % (M,C,A))
    start_time = timeit.default_timer()

    # does this code sound familiar? see previous cell
    pdf = np.arange(1.0, M + 1.0) ** -A
    pdf /= np.sum(pdf)
    hit = sum(pdf[:int(C)])
    
    if verbose:
        print ("\telapsed= %.2f sec" % (timeit.default_timer() - start_time ))
        print ("\tlfu_hit= %.2f " % (hit))
    return hit


## LFU cache hit ratio -- Aproximate solution
def solve_lfu_approx(M=1e7, C=1e3, A=1.0, verbose=True):
    """ 
    Compute LFU cache hit ratio approximating harmonic numbers  
    H_M and H_C with a logarithm log(M) and log(C) ignoring the
    Euler-Mascheroni constant which is o(1). The approximation
    is accurate as M gets latrge
    """
    print("solve_lfu_approx(%g,%s,%s)" % (M,C,A))
    if A == 1:
        h =  np.log(C) / np.log(M)
        if verbose:
            print ("\tlfu_hit= %.2f " % (h))
        return h
    else:
        return -1
    
def solve_lfu_slow(M=1e7, C=1e3, A=1.0, verbose=True):  
    """ 
    Compute the exact LFU cache hit ratio in the simplest way. 
    It is slow for small catalogs M but scales well in memory
    for large M (unlike solve_lfu that can make your machine swap)
    """
    print("solve_lfu_slow(%g,%s,%s)" % (M,C,A))
    start_time = timeit.default_timer()

    sumM = sumC = 0
    #in 2.6 xrange was an iterator and range not
    #in 3.x xrange does not exist and range is an iterator
    # for m in range(1,M+1):
    #better be safe and do a C-style while loop
    m=1
    while m<=M+1:
        sumM += m**-A
        if m==C:
            sumC=sumM
        m += 1   
    hit = sumC/sumM
    
    if verbose:
        print ("\telapsed= %.2f sec" % (timeit.default_timer() - start_time ))
        print ("\tlfu_hit= %.2f " % (hit))
    return hit   

def solve_lru_che(M=1e7, C=1000, A=1, verbose=True):  
    """ 
    Compute the LRU cache hit ratio with the Che approximation
    """ 
    print("solve_lru_che(%g,%s,%s)" % (M,C,A))
    start_time = timeit.default_timer()
               
    pdf = np.arange(1.0, M + 1.0) ** -A
    pdf /= np.sum(pdf)
    contents = range(len(pdf)) 

    def tc_equation(T_C):
        return sum(math.exp(-pdf[j] * T_C) for j in range(len(pdf))) \
            - len(pdf) + C
    
    #Che approximation: consider the random variable of 
    #mean value E[T_C] equal to a constant of value T_C
    T_C = fsolve(tc_equation, x0=C)[0]        
    h_c = [1 - math.exp(-pdf[i] * T_C) for i in contents]
    hit = sum(pdf[i] * h_c[i] for i in range(len(pdf)))
               
    elapsed = timeit.default_timer() - start_time    
    if verbose:
        print ("\tTC = %.2f" % T_C)
        print ("\telapsed= %.2f sec" % (timeit.default_timer() - start_time ))
        print ("\tlru_hit= %.2f " % (hit))           
    return hit           





if __name__ == '__main__':
	#getting started 
	M=1e6
    C=1e3
    alpha=np.linspace(0.6, 1.4, num=10) 
    perc=[]
    for A in alpha:
        pdf = np.arange(1.0, M + 1.0) ** -A
        pdf /= np.sum(pdf)
        #perc.append(np.percentile(pdf,95))
        perc.append(sum(pdf[:int(C)]))
    
    plt.figure()
    plt.plot(alpha, perc)
    plt.show()


	#for small catalogs, solve_lfu is efficient 
	catalog=[1e6, 1e7]
	for M in catalog:
	    solve_lfu(M,C=1e3,A=1.0)
	    solve_lfu_slow(M,C=1e3,A=1.0)
	    solve_lfu_approx(M,C=1e3,A=1.0)


	#for large catalogs, solve_lfu can hang your notebook
	catalog=[1e8, 1e9]
	for M in catalog:
	    #use solve_lfu at your own risk.
	    solve_lfu_slow(M,C=1e3,A=1.0)
	    solve_lfu_approx(M,C=1e3,A=1.0)
	  
	    
	#for very large catalogs, the approximation gets more accurate and
	# since it is also significantly faster, it is OK for a first estimation
	cache=[1e3, 1e4, 1e5]
	catalog=[1e6, 1e7, 1e8, 1e9, 1e10, 1e11, 1e12]
	plt.figure()        
	for C in cache:
	    hit_lfu_approx = []
	    for M in catalog:
	         hit_lfu_approx.append(solve_lfu_approx(M,C,A=1.0,verbose=False))
	    plt.semilogx(catalog, hit_lfu_approx)

	plt.show()

	# compare LRU and LFU
	M=1e7
	solve_lfu(M,C=1e3,A=1.0)
	solve_lfu_approx(M,C=1e3,A=1.0)
	solve_lru_che(M,C=1e3,A=1.0)